# QEWD-Up Life Cycle Events

QEWD-Up provides hooks at various points in the life-cycle of an API, allowing you fine control over the behaviour of your API, whilst allowing you to document your API Life-Cycles in an easy-to-find and easy-to-understand way.

## Life-Cycle Points

The points at which you can intercept the life-cycle of an API depends on whether your application is a monolithic one or implemented as a set of MicroServices.

### Monolithic QEWD-Up Application

Understanding how and why to control the life-cycle of your monolithic application APIs requires an understanding of QEWD's Master Process / Worker Process architecture.  

#### The QEWD Run-Time Environment

This is described [diagrammatically here](https://www.slideshare.net/robtweed/qewdjs-json-web-tokens-microservices) in slides 4 - 40.

In summary, all incoming requests are received by QEWD's Master Process.  In the case of REST API requests, these are received by the embedded Express web server.

The requests are then placed in a queue within the Master Process.  By default, any request payloads are assumed to be JSON-formatted, and the request is passed unchanged from Express to QEWD's queue.

When a new item is added to the queue, the Master Process tries to send it to an available Worker Process.  By default, QEWD-Up configures QEWD with a pool of 2 Worker Processes, but you can modify this to whatever pool size you require in the *config.json* file.

Your API handler module is invoked within the Worker process that receives the incoming request.  Your API handler module has exclusive access to the Worker process, so, unusually for a Node.js environment, your handler module does not need to be concerned about concurrency.  Additionally, the Worker process has synchronous access to the integrated persistent JSON database.

Your API handler module can make use of any 3rd party Node.js modules you wish to use, and, subject to network availability, your API handler module can also send requests to external services and resources.

Once your handler logic is completed, you must invoke QEWD's *finished* function which does two key things:

- it tells QEWD that you have finished with the Worker process;
- it allows you to define the JSON response object that you want to return to the REST client

The *finished* function actually sends the JSON response object to the QEWD Master process, and, on receipt of this message, the Master process returns the Worker process back to the available pool, ready to handle the next queued request.

Finally, the Master process sends the JSON response to the REST Client.  By default, no changes are made to the JSON response by the Master process: QEWD assumes that the REST Client is expecting JSON-formatted.

### Available Life-Cycle Hooks in Chronological Sequence

- **onStarted**: triggered after QEWD's Master process starts and before any APIs can be handled.  It gives you access to QEWD's *this* object and can be used for complex situations, for example where you want to integrate QEWD with a third-party module that provides a turnkey server environment.  This hook is invoked in QEWD's Master process.

- **beforeRouter**: triggered before each incoming API request is routed to its appropriate handler method.  Its usual purpose is to allow you to modify some aspect(s) of the incoming request, for example modifying and/or adding HTTP headers, and/or modifying the request payload.  This hook is invoked in the QEWD Master process, before the incoming request is queued, and applies to **ALL** incoming requests.

- **beforeHandler**: triggered in the QEWD Worker process that handles the incoming request **before** your handler method is invoked. This hook is applied to **all** incoming requests.  Typically it is used to define logic that should be applied to all incoming requests, for example to test user authentication.  The *beforeHandler* can be defined to accept or reject requests based on their headers and/or content, and return an appropriate error response for requests that are rejected.  Rejected messages will **not** invoke your handler method.  Because this hook is invoked in the QEWD Worker process, it has access to the integrated persistent JSON database and/or QEWD Session (if appropriate).

- **afterRouter**: triggered in the QEWD Master process on receipt of the response from the Worker process and before the JSON response is returned to the REST Client.  Typically, this hook is used to modify the response content, eg adding or changing response headers and/or modifying the payload format (eg from JSON to XML).  This hook is applied to **all** responses.  **Note**: if you apply an *afterRouter* hook, then you take on responsibility for sending the response to the REST Client.

